// GETTIME.CPP - main application file for GETTIME.EXE
//
// THIS SOFTWARE IS Copyright 1999 by Bob Frazier and SFT Inc. - All rights reserved
// It has been provided to the public as "copy-left" according to the terms of the
// GNU GENERAL PUBLIC LICENSE at http://www.gnu.org/copyleft/gpl.html
//
// THIS CODE IS PROVIDED 'AS-IS' with NO WARRANTEE WHATSOEVER
// USE IT AT YOUR OWN RISK - Bob Frazier and SFT Inc. CLAIM NO
// RESPONSIBILITY FOR ANY DAMAGES, CONSEQUENTIAL OR OTHERWISE,
// THAT MAY RESULT FROM USE OR ABUSE OF THIS PROGRAM, WHETHER
// IN ITS ORIGINAL OR MODIFIED STATE.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// For a copy of the GNU General Public License, you can visit the
// 'gnu.org' web site at http://www.gnu.org/, or write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// This app runs as a service on NT, or as a background task on Win '9x.  It checks
// the specified IP address for a 'daytime' (TCP 13) service once per minute.  When
// found it updates the current machine's date and time to reflect the settings on the
// server, then goes into a 'sleep' mode.  Every 37 minutes, it attempts to contact
// the server and adjust the date/time as required to match.  The clock will only
// be updated if it differs by more than 5 seconds (to avoid frequent time changes).

// FORMAT: typical formats of 'daytime' return are below (RFC867 is not specific)
// Wednesday, February 10, 1999 9:59:21
// Wed Feb 10 16:19:49 1999
// 51221 99-02-12 20:44:31 00 0 0 169.5 UTC(NIST) *    [see below]
//
// If your server differs from these (or the one below) you can modify the code.
//
// (from http://www.bldrdoc.gov/timefreq/service/nts.htm) time.nist.gov:13
//
// The 'daytime' time format sent by the U.S. government time server(s) is(are):
// JJJJJ YR-MO-DA HH:MM:SS TT L H msADV UTC(NIST) OTM
//
// JJJJJ is the Modified Julian Date (MJD). The MJD is the last
// five digits of the Julian Date, which is simply a count of the
// number of days since January 1, 4713 B.C. To get the Julian Date,
// add 2.4 million to the MJD.
//
// YR-MO-DA is the date. It shows the last two digits of the year,
// the month, and the current day of month
//
// HH:MM:SS is the time in hours, minutes, and seconds. The time is
// always sent as Coordinated Universal Time (UTC). An offset needs
// to be applied to UTC to obtain local time. For example, Mountain
// Time in the U. S. is 7 hours behind UTC during Standard Time, and
// 6 hours behind UTC during Daylight Saving Time
//
// TT is a two digit code (00 to 99) that indicates whether the United
// States is on Standard Time (ST) or Daylight Saving Time (DST). It
// also indicates when ST or DST is approaching. This code is set to 00
// when ST is in effect, or to 50 when DST is in effect. During the
// month in which the time change actually occurs, this number will
// decrement every day until the change occurs. For example, during the
// month of October, the U.S. changes from DST to ST. On October 1, the
// number will change from 50 to the actual number of days until the
// time change. It will decrement by 1 every day until the change occurs
// at 2 a.m. local time when the value is 1. Likewise, the spring change
// is at 2 a.m. local time when the value reaches 51
//
// L is a one-digit code that indicates whether a leap second will be
// added or subtracted at midnight on the last day of the current month.
// If the code is 0, no leap second will occur this month. If the code is
// 1, a positive leap second will be added at the end of the month. This
// means that the last minute of the month will contain 61 seconds instead
// of 60. If the code is 2, a second will be deleted on the last day of the
// month. Leap seconds occur at a rate of about one per year. They are used
// to correct for irregularity in the earth's rotation. The correction is
// made just before midnight UTC (not local time).
//
// H is a health digit that indicates the health of the server. If H=0,
// the server is healthly. If H=1, then the server is operating properly
// but its time may be in error by up to 5 seconds. This state should change
// to fully healthy within 10 minutes. If H=2, then the server is operating
// properly but its time is known to be wrong by more than 5 seconds. If H=4,
// then a hardware or software failure has occurred and the amount of the
// time error is unknown. 
//
// msADV displays the number of milliseconds that NIST advances the time code
// to partially compensate for network delays. The advance is currently set to
// 50.0 milliseconds.
//
// The label UTC(NIST) is contained in every time code. It indicates that you
// are receiving Coordinated Universal Time (UTC) from the National Institute
// of Standards and Technology (NIST).
//
// OTM (on-time marker) is an asterisk (*). The time values sent by the time
// code refer to the arrival time of the OTM. In other words, if the time code
// says it is 12:45:45, this means it is 12:45:45 when the OTM arrives.
//



#include "stdafx.h"
#include "gettime.h"
#include "DateTimeThread.h"


#define STR_RUNSERVICES_KEY "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\RunServices"
#define STR_APPLICATION_KEY "SOFTWARE\\Stewart~Frazier Tools\\GetTime"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// global variables
SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;
BOOL                    bDebug = FALSE;
BOOL                    bIsNT = FALSE;
DWORD                   dwLastErr = 0;

static HANDLE hServerStopEvent = NULL;

static CString csLastError;
static DWORD dwThreadId = 0;

SERVICE_TABLE_ENTRY dispatchTable[] =
{
  { SZSERVICENAME, (LPSERVICE_MAIN_FUNCTION)service_main },
  { NULL, NULL }
};


/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CGetTimeApp theApp;

using namespace std;  // not sure why... it's an MFC thing I guess

BEGIN_MESSAGE_MAP(CGetTimeApp, CWinApp)
	//{{AFX_MSG_MAP(CGetTimeApp)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CWnd CGetTimeApp::m_wndNULL;  // creates a null-handled window object

CGetTimeApp::CGetTimeApp()
{
  m_pMainWnd = &m_wndNULL;
}

CGetTimeApp::~CGetTimeApp()
{
}

BOOL CGetTimeApp::MyMessageLoop() // adapted from CWinThread::Run()
{
  // an 'inside-out' message loop - calls message pump *ONLY* when there
  // are things to process.  Calls the idle loop one time only.

  static BOOL bIdle = TRUE;
	static LONG lIdleCount = 0;

  DWORD dwStartTick = GetTickCount();

	// acquire and dispatch messages until a WM_QUIT message is received.
	// phase1: check to see if we can do idle work

	// phase2: pump messages while available
	while(::PeekMessage(&m_msgCur, NULL, NULL, NULL, PM_NOREMOVE)
        && (GetTickCount() - dwStartTick) < 500) // 1/2 sec of messages max
	{
		// pump message, but quit on WM_QUIT
		if(!PumpMessage())
    {
      return(TRUE);  // assumes a 'quit' message was received.
//			return ExitInstance();

    }

		// reset "no idle" state after pumping "normal" message
		if(IsIdleMessage(&m_msgCur))
		{
			bIdle = TRUE;
			lIdleCount = 0;
		}
	}

	if(bIdle && (GetTickCount() - dwStartTick) < 500)  // assumes no message
	{
		// call OnIdle while in bIdle state
		if(!OnIdle(lIdleCount++))
			bIdle = FALSE; // assume "no idle" state
	}

  return(FALSE);
}

/////////////////////////////////////////////////////////////////////////////
// 'MAIN' proc for application entry (oh, boy!)

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
  int nRetCode = 0; // default 'ok'
  BOOL bShowHelp = TRUE;

  // initialize MFC and print and error on failure
  if(!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
  {
    // request access to desktop

    ::MessageBox(NULL, _T("Fatal Error: MFC initialization failed"),
                 _T("* GetTime Service *"),
                 MB_OK | MB_ICONHAND | MB_SETFOREGROUND);

    return(1);
  }

  theApp.SetRegistryKey("Stewart~Frazier Tools");

  TCHAR tbuf[MAX_PATH * 2]=_T("");

  GetConsoleTitle(tbuf, sizeof(tbuf));
  wsprintf(tbuf + lstrlen(tbuf), " %08xH", theApp.m_nThreadID);

  HWND hConsole = FindWindow(NULL, tbuf);
  // find the console window - I may need it...


  // if I'm running under Win '9x, the service registration functionality is
  // completely different

  OSVERSIONINFO ver;

  memset(&ver, 0, sizeof(ver));
  ver.dwOSVersionInfoSize = sizeof(ver);

  if(!GetVersionEx(&ver) ||
     ver.dwPlatformId == VER_PLATFORM_WIN32s ||
     ver.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
  {
    bIsNT = FALSE;
  }
  else
  {
    ASSERT(ver.dwPlatformId == VER_PLATFORM_WIN32_NT);

    bIsNT = TRUE;
  }

  if(argc > 1)
  {
    if(argv[1][0] != '/' && argv[1][0] != '-')
    {
      // invalid switch character

      _tprintf(_T("?Invalid parameter - %s\n"), argv[1]);

      nRetCode = 2;  // "bad argument"
    }
    else if(!stricmp(argv[1] + 1, "start") ||
            !stricmp(argv[1] + 1, "install"))
    {
      if(bIsNT)
      {
        CmdInstallService(!stricmp(argv[1] + 1, "start"));
      }
      else
      {
        *tbuf = 0;
        GetModuleFileName(NULL, tbuf, sizeof(tbuf));

        if(!*tbuf)
        {
          _tprintf(_T("Unable to get module file name - %s\n"), GetLastErrorText());

          nRetCode = 3;  // "failed to install"
        }
        else
        {
          // for Win '9x, add to 'RunServices' with argument '-95'
          // and then RUN the thing if the user said '-start'.

          HKEY hKey;
          DWORD dwDisp;

          CString csTemp = "\"" + (CString)tbuf + "\" -95";

          if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, STR_RUNSERVICES_KEY,
                            0, NULL, REG_OPTION_NON_VOLATILE,
                            KEY_ALL_ACCESS, NULL,
                            &hKey, &dwDisp)
             == ERROR_SUCCESS)
          {
            if(RegSetValueEx(hKey, "GetTime", 0, REG_SZ,
                             (BYTE *)(LPCTSTR)csTemp,
                             (csTemp.GetLength() + 1) * sizeof(csTemp[0]))
               == ERROR_SUCCESS)
            {
              STARTUPINFO si;
              PROCESS_INFORMATION pi;

              memset(&si, 0, sizeof(si));
              si.cb = sizeof(si);
              si.wShowWindow = SW_HIDE;
              si.dwFlags = STARTF_USESHOWWINDOW;

              memset(&pi, 0, sizeof(pi));

              CString csPath = tbuf;
              int i1 = csPath.ReverseFind('\\');
              if(i1 > 3)
                csPath = csPath.Left(i1);
              else
                csPath = csPath.Left(i1 + 1);

              _tprintf(_T("Starting %s..."), (LPCSTR)tbuf);

              if(!stricmp(argv[1] + 1, "start") &&
                 !CreateProcess(NULL, (LPSTR)(LPCSTR)csTemp, // tbuf, "-95",
                                NULL, NULL, FALSE,
                                CREATE_NEW_CONSOLE | NORMAL_PRIORITY_CLASS, // DETACHED_PROCESS | NORMAL_PRIORITY_CLASS,
                                NULL, csPath, &si, &pi))
              {
                _tprintf(_T("\nUnable to start service - %s\n"), (LPCTSTR)GetLastErrorText());

                nRetCode = 3;
              }
              else
              {
                _tprintf(_T(".. Complete!\n"));
              }
            }
            else
            {
              _tprintf(_T("Unable to create value for registry key - %s\n"),
                       GetLastErrorText());

              nRetCode = 3;  // "failed to install"
            }

            RegCloseKey(hKey);
          }
          else
          {
            _tprintf(_T("Unable to create registry key - %s\n"), GetLastErrorText());

            nRetCode = 3;  // "failed to install"
          }
        }
      }

      // using arguments provided to this application, assign the
      // appropriate registry values.

      HKEY hKey;
      DWORD dwDisp;

      if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY,
                        0, NULL, REG_OPTION_NON_VOLATILE,
                        KEY_ALL_ACCESS, NULL, 
                        &hKey, &dwDisp)
         == ERROR_SUCCESS)
      {
        // add strings for each argument, each separated by white space

        int i1 = 2;  // total # of arguments thus far, index for NEXT one

        if(argc > i1
           && (argv[i1][0] == '-' || argv[i1][0] == '/')
           && !strnicmp(argv[i1] + 1, "SOCKS=", 6))
        {
          // specifying a SOCKS server

          RegSetValueEx(hKey, "SOCKS", 0, REG_SZ,
                        (const BYTE *)(argv[i1] + 7),
                        (strlen(argv[i1] + 7) + 1) * sizeof(**argv));

          i1++;  // advance to next arg
        }

        if(argc > i1) // an IP address
        {

          RegSetValueEx(hKey, "Address", 0, REG_SZ,
                        (const BYTE *)argv[i1],
                        (strlen(argv[i1]) + 1) * sizeof(**argv));

          i1++;
        }

        if(argc > i1) // a time zone or adjustment
        {
          RegSetValueEx(hKey, "TimeZone", 0, REG_SZ,
                        (const BYTE *)argv[i1],
                        (strlen(argv[i1]) + 1) * sizeof(**argv));

          i1++;
        }

        RegCloseKey(hKey);
      }

      bShowHelp = FALSE;
    }
    else if(!stricmp(argv[1] + 1, "stop") ||
            !stricmp(argv[1] + 1, "uninstall"))
    {
      if(bIsNT)
      {
        CmdRemoveService(!stricmp(argv[1] + 1, "uninstall"));
      }
      else
      {
        // for Win '9x, remove from 'RunServices' and then
        // tell the app to stop.  Use the registry to find
        // the right event and signal it.
        // This should be SZGLOBALEVENTNAME

        HKEY hKey;

        if(RegOpenKey(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY, &hKey)
           == ERROR_SUCCESS)
        {
          DWORD dwProcessID, dwThreadID, dwType, cb1;

          cb1 = sizeof(dwProcessID);

          if(RegQueryValueEx(hKey, "ProcessID", NULL, &dwType, (BYTE *)&dwProcessID, &cb1)
             == ERROR_SUCCESS)
          {
            HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwProcessID);

            if(!hProcess)
            {
              _tprintf(_T("WARNING - unable to query process ID %08x\n"), dwProcessID);
            }

            ASSERT(cb1 == sizeof(dwProcessID));

            cb1 = sizeof(dwProcessID);
            if(RegQueryValueEx(hKey, "ThreadID", NULL, &dwType, (BYTE *)&dwThreadID, &cb1)
               == ERROR_SUCCESS)
            {
              ASSERT(cb1 == sizeof(dwThreadID));

              _tprintf(_T("Stopping gettime..."));

//              ::PostThreadMessage(dwProcessID, WM_QUIT, 0, 0);
              ::PostThreadMessage(dwThreadID, WM_QUIT, 0, 0);


              // wait for it to go away

              cb1 = STILL_ACTIVE;  // as a flag

              do
              {
                Sleep(500);  // always wait .5 seconds

                _tprintf(_T("."));

              } while(hProcess && 
                      GetExitCodeProcess(hProcess, &cb1) &&
                      cb1 == STILL_ACTIVE);

              if(cb1 != STILL_ACTIVE)
                _tprintf(_T(" STOPPED! - exit code %d"), cb1);
              else
                _tprintf(_T(" stopped?\n"));
            }
            else
            {
              _tprintf(_T("ERROR - unable to get main thread ID for GetTime from registry\n        Cannot stop GetTime service\n"));
            }

            if(hProcess)
              CloseHandle(hProcess);
          }
          else
          {
            _tprintf(_T("Unable to stop GetTime process - no Process ID registered\n"));
          }

          RegCloseKey(hKey);
        }
        else
        {
          _tprintf(_T("The GetTime service was not installed\n"));
        }

        if(RegOpenKey(HKEY_LOCAL_MACHINE, STR_RUNSERVICES_KEY, &hKey)
           == ERROR_SUCCESS)
        {
          if(RegDeleteValue(hKey, "GetTime")
             != ERROR_SUCCESS)
          {
            _tprintf(_T("Unable to remove registry entries - %s\n"),
                     GetLastErrorText());
          }

          RegCloseKey(hKey);
        }
      }

      if(!stricmp(argv[1] + 1, "uninstall") &&
         RegDeleteKey(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY)
         != ERROR_SUCCESS)
      {
        _tprintf(_T("Unable to remove registry entries - %s\n"),
                 GetLastErrorText());
      }

      bShowHelp = FALSE;
    }
    else if(!stricmp(argv[1] + 1, "debug") || !stricmp(argv[1] + 1, "d"))
    {
      bDebug = TRUE;
      bShowHelp = FALSE;

      CmdDebugService(argc, argv);
    }
    else if(!stricmp(argv[1] + 1, "95"))
    {
      // free the console window

      bDebug = FALSE;  // this has special properties associated with it...

      HKEY hKey;
      DWORD dwDisp;

      if(RegCreateKeyEx(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY,
                        0, NULL, REG_OPTION_NON_VOLATILE,
                        KEY_ALL_ACCESS, NULL, 
                        &hKey, &dwDisp)
         == ERROR_SUCCESS)
      {
        DWORD dwProcessID = GetCurrentProcessId();

        if(RegSetValueEx(hKey, "ThreadID", 0, REG_DWORD,
                         (BYTE *)&theApp.m_nThreadID, sizeof(theApp.m_nThreadID))
           != ERROR_SUCCESS || 
           RegSetValueEx(hKey, "ProcessID", 0, REG_DWORD,
                         (BYTE *)&dwProcessID, sizeof(dwProcessID))
           != ERROR_SUCCESS)
        {
          // start fail

          RegDeleteValue(hKey, "ThreadID");
          RegDeleteValue(hKey, "ProcessID");

          RegCloseKey(hKey);

          _tprintf(_T("GetTime - Unable to set process/thread ID's in registry - %s\n"),
                   GetLastErrorText());
        }
        else
        {
          RegCloseKey(hKey);

          // ensure window is hidden first...

          ShowWindow(hConsole, SW_HIDE);
          FreeConsole();  // byby!

          CmdDebugService(argc, argv);  // this really runs it

          AddToMessageLog("GetTime application terminating", EVENTLOG_INFORMATION_TYPE);

          // upon return, I must delete this registry entry - but ignore errors doing it

          if(RegOpenKey(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY,
                        &hKey) == ERROR_SUCCESS)
          {
            if(RegDeleteValue(hKey, "ProcessID")
               != ERROR_SUCCESS)
            {
              ASSERT(0);  // for debugging only

              TRACE("UNABLE TO DELETE 'ProcessID' - %s\r\n", GetLastErrorText());

              nRetCode = 4;

              AddToMessageLog("UNABLE TO DELETE 'ProcessID' key");
            }

            RegCloseKey(hKey);
          }

          bShowHelp = FALSE;
        }
      }
    }
    else if(stricmp(argv[1] + 1, "?") &&
            stricmp(argv[1] + 1, "HELP") &&
            stricmp(argv[1] + 1, "h"))
    {
      _tprintf(_T("?Invalid switch - %s\n"), argv[1]);

      nRetCode = 2;  // "bad argument"
    }
  }
  else if(bIsNT)
  {
    _tprintf(_T("Use gettime -? for help.\nAttempting to register with SCM... (press CTRL+C to abort)"));

    if(!StartServiceCtrlDispatcher(dispatchTable))  // fails because why?
    {
      AddToMessageLog(TEXT("StartServiceCtrlDispatcher failed."));

      _tprintf(_T(" FAILED!\n\n"));

      bShowHelp = TRUE;
    }
    else
    {
      bShowHelp = FALSE;
    }
  }

  if(bShowHelp)
  {
    // if I get here, print the 'help' horse

    CString csTemp;

    csTemp.LoadString(IDS_COPYRIGHT);
    _tprintf(csTemp);
    csTemp.LoadString(IDS_HELP1);
    _tprintf(csTemp);
    csTemp.LoadString(IDS_HELP2);
    _tprintf(csTemp);

    _tprintf(_T("Press <ENTER> to continue:"));
    getc(stdin);
    _tprintf("\n");

    csTemp.LoadString(IDS_HELP3);
    _tprintf(csTemp);

    _tprintf(_T("Press <ENTER> to continue:"));
    getc(stdin);

    _tprintf(_T("\n\n"));
  }

  return nRetCode;
}

void CmdInstallService(BOOL bStartFlag)
{
  SC_HANDLE   schService;
  SC_HANDLE   schSCManager;

  TCHAR szPath[512];

  if(GetModuleFileName( NULL, szPath, 512 ) == 0)
  {
    _tprintf(_T("Unable to install %s - %s\n"), SZSERVICEDISPLAYNAME,
             GetLastErrorText());
    return;
  }

  schSCManager = OpenSCManager(NULL,                   // machine (NULL == local)
                               NULL,                   // database (NULL == default)
                               SC_MANAGER_ALL_ACCESS); // access required

  if(schSCManager)
  {
    schService = CreateService(schSCManager,               // SCManager database
                               SZSERVICENAME,              // name of service
                               SZSERVICEDISPLAYNAME,       // name to display
                               SERVICE_ALL_ACCESS,         // desired access
                               SERVICE_WIN32_OWN_PROCESS   // service type
                               | SERVICE_INTERACTIVE_PROCESS, // allows it to interact with desktop
//                               SERVICE_DEMAND_START,       // start type
                               SERVICE_AUTO_START,         // I want automatic startup
                               SERVICE_ERROR_NORMAL,       // error control type
                               szPath,                     // service's binary
                               NULL,                       // no load ordering group
                               NULL,                       // no tag identifier
                               SZDEPENDENCIES,             // dependencies
                               NULL,                       // LocalSystem account
                               NULL);                      // no password


    if(!schService && GetLastError() == ERROR_SERVICE_EXISTS)
    {
      // service already exists? - attempt to open

      schService = OpenService(schSCManager, SZSERVICENAME, SERVICE_ALL_ACCESS);
    }

    if( schService)
    {
      if(bStartFlag)
      {
        _tprintf(_T("%s installed.  Starting..."), SZSERVICEDISPLAYNAME );

        LPCTSTR lp1 = szPath;

        if(!StartService(schService, 1, &lp1))
        {
          _tprintf(_T(" %s\n"), (LPCTSTR)GetLastErrorText());
        }
        else
        {
          _tprintf(_T(" Complete!\n"));
        }
      }
      else
      {
        _tprintf(_T("%s installed (not started)."), SZSERVICEDISPLAYNAME );
      }

      CloseServiceHandle(schService);
    }
    else
    {
      _tprintf(_T("CreateService failed - %s\n"), GetLastErrorText());
    }

    CloseServiceHandle(schSCManager);
  }
  else
  {
    _tprintf(_T("OpenSCManager failed - %s\n"), GetLastErrorText());
  }
}

void CmdRemoveService(BOOL bUninstall)
{
  SC_HANDLE   schService;
  SC_HANDLE   schSCManager;

  schSCManager = OpenSCManager(NULL,                   // machine (NULL == local)
                               NULL,                   // database (NULL == default)
                               SC_MANAGER_ALL_ACCESS); // access required

  if(schSCManager)
  {
    schService = OpenService(schSCManager, SZSERVICENAME, SERVICE_ALL_ACCESS);

    if(schService)
    {
      // try to stop the service
      if(ControlService(schService, SERVICE_CONTROL_STOP, &ssStatus))
      {
        _tprintf(_T("Stopping %s."), SZSERVICEDISPLAYNAME);
        Sleep( 1000 );

        while(QueryServiceStatus( schService, &ssStatus ))
        {
          if(ssStatus.dwCurrentState == SERVICE_STOP_PENDING)
          {
            _tprintf(_T("."));
            Sleep( 1000 );
          }
          else
            break;
        }

        if(ssStatus.dwCurrentState == SERVICE_STOPPED)
          _tprintf(_T("\n%s stopped.\n"), SZSERVICEDISPLAYNAME );
        else
        {
          dwLastErr = GetLastError();
          if(dwLastErr)
          {
            _tprintf(_T("\n%s failed to stop - %s.\n"),
                     SZSERVICEDISPLAYNAME, GetLastErrorText() );
          }
          else
          {
            _tprintf(_T("\n%s failed to stop.\n"), SZSERVICEDISPLAYNAME );
          }
        }
      }
      else if((dwLastErr = GetLastError()) != ERROR_SERVICE_NOT_ACTIVE)
      {
        _tprintf(_T("Unable to stop service - %s\n"), GetLastErrorText());
      }

      if(bUninstall)
      {
        // now remove the service
        if(DeleteService(schService))
        {
          _tprintf(_T("%s removed.\n"), SZSERVICEDISPLAYNAME);
        }
        else
        {
          _tprintf(_T("DeleteService failed - %s\n"), GetLastErrorText());
        }
      }


      CloseServiceHandle(schService);
    }
    else
      _tprintf(_T("OpenService failed - %s\n"), GetLastErrorText());

    CloseServiceHandle(schSCManager);
  }
  else
    _tprintf(_T("OpenSCManager failed - %s\n"), GetLastErrorText());
}

void CmdDebugService(int argc0, char ** argv0)
{
  if(!bDebug)  // this is how I invoke it in "-95" mode!!!!
  {
    int argc = 1;
    const TCHAR *argv[4];
    TCHAR szAppName[MAX_PATH * 2];

    GetModuleFileName(NULL, szAppName, sizeof(szAppName));
    argv[0] = szAppName;

    HKEY hKey;
    DWORD cb1, dwType;

    if(RegOpenKey(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY,
                  &hKey)
       == ERROR_SUCCESS)
    {
      CString csTemp, csSOCKS, csAddr, csTZ;

      cb1 = 256; // max size

      if(RegQueryValueEx(hKey, "SOCKS", NULL, &dwType,
                         (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
         == ERROR_SUCCESS)
      {
        csTemp.ReleaseBuffer(cb1);
        csTemp.ReleaseBuffer(-1);  // trims excess junk
      }
      else
      {
        csTemp.ReleaseBuffer(0);
      }

      if(csTemp.GetLength())
      {
        csSOCKS = "-SOCKS=" + csTemp;
        argv[argc++] = (LPCTSTR)csSOCKS;
      }

      cb1 = 256; // max size

      if(RegQueryValueEx(hKey, "Address", NULL, &dwType,
                         (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
         == ERROR_SUCCESS)
      {
        csTemp.ReleaseBuffer(cb1);
        csTemp.ReleaseBuffer(-1);  // trims excess junk
      }
      else
      {
        csTemp.ReleaseBuffer(0);
      }

      if(csTemp.GetLength())
      {
        csAddr = csTemp;
        argv[argc++] = (LPCTSTR)csAddr;
      }

      cb1 = 256; // max size

      if(RegQueryValueEx(hKey, "TimeZone", NULL, &dwType,
                         (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
         == ERROR_SUCCESS)
      {
        csTemp.ReleaseBuffer(cb1);
        csTemp.ReleaseBuffer(-1);  // trims excess junk
      }
      else
      {
        csTemp.ReleaseBuffer(0);
      }

      if(csTemp.GetLength())
      {
        csTZ = csTemp;
        argv[argc++] = (LPCTSTR)csTZ;
      }

      RegCloseKey(hKey);

      ServiceStart( argc, (char **)argv );
    }
    else
    {
      ServiceStart(argc0, argv0); // make it run anyway, with THAT argument...
    }
  }
  else
  {
    DWORD dwArgc;
    LPTSTR *lpszArgv;

  #ifdef UNICODE
    lpszArgv = CommandLineToArgvW(GetCommandLineW(), &(dwArgc));
  #else
    dwArgc   = (DWORD)argc0;
    lpszArgv = argv0;
  #endif

    _tprintf(_T("Debugging %s.\n"), SZSERVICEDISPLAYNAME);

    if(bDebug)
    {
      SetConsoleCtrlHandler( ControlHandler, TRUE );
    }

    // the first arg is always "not wanted" so get rid of it...

    TCHAR **argv2 = new TCHAR * [dwArgc];
    if(argv2)
    {
      int i1;
      argv2[0] = lpszArgv[0];       // leave THIS one the same

      for(i1=2; i1 < dwArgc; i1++)  // and get rid of lpszArgv[1]
      {
        argv2[i1 - 1] = lpszArgv[i1];
      }

      ServiceStart(dwArgc - 1, argv2);  // ideal method
    }
    else
    {
      ServiceStart(dwArgc - 1, lpszArgv + 1);  // "fudge" it
    }
  }
}

BOOL WINAPI ControlHandler ( DWORD dwCtrlType )
{
  switch( dwCtrlType )
  {
    case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
    case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode

      _tprintf(_T("Stopping %s.\n"), SZSERVICEDISPLAYNAME);

      ServiceStop();

      return TRUE;
  }

  return FALSE;
}

CString GetLastErrorText()
{
  CString csRval;

  DWORD dwRet;
  LPTSTR lpszTemp = NULL;

  dwLastErr = GetLastError();

  dwRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER
                        | FORMAT_MESSAGE_FROM_SYSTEM
                        | FORMAT_MESSAGE_ARGUMENT_ARRAY,
                        NULL,
                        dwLastErr,
                        LANG_NEUTRAL,
                        (LPTSTR)&lpszTemp,
                        0,
                        NULL);

  if(dwRet)
  {
    lpszTemp[strlen(lpszTemp)-2] = _T('\0');

    csRval.Format("%s (%0x%x)", lpszTemp, dwLastErr);
  }

  if(lpszTemp)
    LocalFree((HLOCAL)lpszTemp);

  return(csRval);
}



/////////////////////////////////////////////////////////////////////////////
// ** SERVICE CONTROL FUNCTIONS **

void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{
  // register our service control handler:

  sshStatusHandle = RegisterServiceCtrlHandler( SZSERVICENAME, service_ctrl);

  if(!sshStatusHandle)
  {
    AddToMessageLog("RegisterServiceCtrlHandler() failed in 'service_main()'");
    return;
  }  

  // SERVICE_STATUS members that don't change in example

  ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  ssStatus.dwServiceSpecificExitCode = 0;


  // report the status to the service control manager.

  if(ReportStatusToSCMgr(SERVICE_START_PENDING, // service state
                         NO_ERROR,              // exit code
                         3000))                 // wait hint
  {
//    if(dwArgc > 1) // arguments passed during StartService() call?
//    {
//      ServiceStart( dwArgc, lpszArgv );
//    }
//    else
    {
      int argc = 1;
      const TCHAR *argv[4];
      TCHAR szAppName[MAX_PATH * 2];

      GetModuleFileName(NULL, szAppName, sizeof(szAppName));
      argv[0] = szAppName;

      HKEY hKey;
      DWORD cb1, dwType;

      if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, STR_APPLICATION_KEY,
                      0, KEY_QUERY_VALUE, &hKey)
         == ERROR_SUCCESS)
      {
        CString csTemp, csSOCKS, csAddr, csTZ;

        cb1 = 256; // max size

        if(RegQueryValueEx(hKey, "SOCKS", NULL, &dwType,
                           (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
           == ERROR_SUCCESS)
        {
          csTemp.ReleaseBuffer(cb1);
          csTemp.ReleaseBuffer(-1);  // trims excess junk
        }
        else
        {
          csTemp.ReleaseBuffer(0);
        }

        if(csTemp.GetLength())
        {
          csSOCKS = "-SOCKS=" + csTemp;
          argv[argc++] = (LPCTSTR)csSOCKS;
        }

        cb1 = 256; // max size

        if(RegQueryValueEx(hKey, "Address", NULL, &dwType,
                           (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
           == ERROR_SUCCESS)
        {
          csTemp.ReleaseBuffer(cb1);
          csTemp.ReleaseBuffer(-1);  // trims excess junk
        }
        else
        {
          csTemp.ReleaseBuffer(0);
        }

        if(csTemp.GetLength())
        {
          csAddr = csTemp;
          argv[argc++] = (LPCTSTR)csAddr;
        }

        cb1 = 256; // max size

        if(RegQueryValueEx(hKey, "TimeZone", NULL, &dwType,
                           (BYTE *)csTemp.GetBufferSetLength(cb1), &cb1)
           == ERROR_SUCCESS)
        {
          csTemp.ReleaseBuffer(cb1);
          csTemp.ReleaseBuffer(-1);  // trims excess junk
        }
        else
        {
          csTemp.ReleaseBuffer(0);
        }

        if(csTemp.GetLength())
        {
          csTZ = csTemp;
          argv[argc++] = (LPCTSTR)csTZ;
        }

        RegCloseKey(hKey);

//        AddToMessageLog(_T("TEMPORARY - GetTime - ServiceStart(")
//                        + csSOCKS + _T(",")
//                        + csAddr + _T(",")
//                        + csTZ + _T(")"));
      
        ServiceStart( argc, (char **)argv );
      }
      else
      {
        AddToMessageLog(_T("GetTime() - not properly installed - unable to load arguments from registry"));

        ServiceStart( dwArgc, lpszArgv );
      }
    }
  }
  else
  {
    AddToMessageLog("SCM not responding - startup failed in 'service_main()'");
  }


  // HERE, when it ends, the function 'ServiceStart' returns.
  // try to report the stopped status to the service control manager.

  if(sshStatusHandle)
  {
    (VOID)ReportStatusToSCMgr(SERVICE_STOPPED,
                              dwLastErr = GetLastError(),
                              0);
  }

  return;
}

VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
  // SCM calls this via 'ControlService()'

  // Handle the requested control code.
  //
  switch(dwCtrlCode)
  {
    // Stop the service.
    //
    // SERVICE_STOP_PENDING should be reported before
    // setting the Stop Event - hServerStopEvent - in
    // ServiceStop().  This avoids a race condition
    // which may result in a 1053 - The Service did not respond...
    // error.

    case SERVICE_CONTROL_STOP:
      ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
      ServiceStop();
      return;

    // Update the service status.
    //
    case SERVICE_CONTROL_INTERROGATE:

      // TODO:  go out and refresh the date/time *right fucking now*

      break;

    // invalid control code
    //
    default:
      break;

  }

  ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
}

BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint)
{
  static DWORD dwCheckPoint = 1;
  BOOL fResult = TRUE;

  if(!bDebug && bIsNT) // when debugging we don't report to the SCM
  {
    if(dwCurrentState == SERVICE_START_PENDING)
    {
      ssStatus.dwControlsAccepted = 0;
    }
    else
    {
      ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    }

    ssStatus.dwCurrentState = dwCurrentState;
    ssStatus.dwWin32ExitCode = dwWin32ExitCode;
    ssStatus.dwWaitHint = dwWaitHint;

    if((dwCurrentState == SERVICE_RUNNING) ||
       (dwCurrentState == SERVICE_STOPPED))
    {
      ssStatus.dwCheckPoint = 0;
    }
    else
    {
      ssStatus.dwCheckPoint = dwCheckPoint++;
    }

    // Report the status of the service to the service control manager.
    //
    if(!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus)))
    {
      AddToMessageLog(_T("SetServiceStatus"));
    }
  }

  return fResult;
}

VOID AddToMessageLog(LPCTSTR lpszMsg, WORD wType /* = EVENTLOG_ERROR_TYPE */)
{
  TCHAR   szMsg[256];
  HANDLE  hEventSource;
  LPCTSTR lpszStrings[2];

  if(!dwLastErr)
    dwLastErr = GetLastError();

  if(dwLastErr)
    _stprintf(szMsg, _T("%s error: %d"), SZSERVICENAME, dwLastErr);
  else
    strcpy(szMsg, SZSERVICENAME);

  if(!bDebug && bIsNT)
  {
    // Use event logging to log the error.

    hEventSource = RegisterEventSource(NULL, SZSERVICENAME);

    lpszStrings[0] = szMsg;
    lpszStrings[1] = lpszMsg;

    if(hEventSource != NULL)
    {
      ReportEvent(hEventSource,          // handle of event source
                  wType, // EVENTLOG_ERROR_TYPE,   // event type
                  0,                     // event category
                  0,                     // event ID
                  NULL,                  // current user's SID
                  2,                     // strings in lpszStrings
                  0,                     // no bytes of raw data
                  (LPCTSTR *)lpszStrings,// array of error strings
                  NULL);                 // no raw data

      (VOID)DeregisterEventSource(hEventSource);
    }
  }
  else if(bDebug)
  {
    _tprintf(_T("%s\n%s\n"), szMsg, lpszMsg);
  }
  else // if(!bIsNT)    windows '95 version
  {
    // FOR WINDOWS '9x ALL EVENTS ARE WRITTEN TO 'SFTEVENT.LOG' WHICH SUBSTITUTES FOR
    // THE WINDOWS NT EVENT LOG AS FAR AS THIS (AND OTHER) SFT SERVICE APPLICATIONS
    // ARE CONCERNED.

    CString csTemp, csMsg;
    GetWindowsDirectory(csTemp.GetBufferSetLength(MAX_PATH), MAX_PATH);
    csTemp.ReleaseBuffer(-1);

    if(csTemp.GetLength() && csTemp[csTemp.GetLength() - 1] != '\\')
      csTemp += '\\';

    csTemp += "SFTEVENT.LOG";
    LPCSTR szType = "[UNKNOWN ERROR TYPE]";
    if(wType == EVENTLOG_ERROR_TYPE)
      szType = "** ERROR **";
    else if(wType == EVENTLOG_WARNING_TYPE)
      szType = "** WARNING **";
    else if(wType == EVENTLOG_INFORMATION_TYPE)
      szType = "** INFORMATION **";

    SYSTEMTIME tm;
    GetSystemTime(&tm);
    csMsg.Format("%s %s %4d-%2d-%2d %02d:%02d:%02d GMT\r\n%s\r\n%s\r\n",
                 (LPCSTR)SZSERVICENAME, szType,
                 tm.wYear, tm.wMonth, tm.wDay,
                 tm.wHour, tm.wMinute, tm.wSecond,
                 szMsg, lpszMsg);

    csMsg += "*************************************************************************\r\n";

    HANDLE hFile = CreateFile(csTemp, GENERIC_READ | GENERIC_WRITE, 0, NULL,
                              OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if(hFile == INVALID_HANDLE_VALUE)
    {
      ::MessageBox(NULL, csMsg, "** SFT GETTIME ERROR **",
                   MB_OK | MB_ICONHAND | MB_SETFOREGROUND);
    }
    else
    {
      DWORD cb1;

      SetFilePointer(hFile, 0, NULL, FILE_END);  // go to the end

      if(!WriteFile(hFile, csMsg, csMsg.GetLength(), &cb1, NULL) ||
         cb1 != csMsg.GetLength())
      {
        ::MessageBox(NULL, csMsg, "** SFT GETTIME LOG WRITE ERROR **",
                     MB_OK | MB_ICONHAND | MB_SETFOREGROUND);
      }

      CloseHandle(hFile);
    }
  }
}


/////////////////////////////////////////////////////////////////////////////
// ** SERVICE 'main' FUNCTIONALITY **

VOID ServiceStart(DWORD dwArgc, LPTSTR *lpszArgv)
{
  SECURITY_DESCRIPTOR sdALL;
  SECURITY_ATTRIBUTES saALL;

  memset(&saALL, 0, sizeof(saALL));

  saALL.nLength = sizeof(saALL);

  if(bIsNT)
  {
    if(InitializeSecurityDescriptor(&sdALL, SECURITY_DESCRIPTOR_REVISION))
    {
      // assign access to 'all users'
      SetSecurityDescriptorDacl(&sdALL, TRUE, NULL, FALSE);

      saALL.lpSecurityDescriptor = &sdALL;
    }
  }

  CEvent event(FALSE, FALSE, SZGLOBALEVENTNAME, &saALL);


  // find the argument with the TCP address in it

  CString csAddr, csSocks, csTZ="";
  int i1;

  for(i1=1; i1 < (int)dwArgc; i1++)
  {
    if(lpszArgv[i1][0] != '/' && lpszArgv[i1][0] != '-') // skip 'switches'
    {
      break;
    }
    else if(!strnicmp(lpszArgv[i1] + 1, "SOCKS=", 6))
    {
      csSocks = lpszArgv[i1] + 7;

      csSocks.TrimRight();
      csSocks.TrimLeft();
    }
  }

  dwThreadId = GetCurrentThreadId();

  if(i1 < (int)dwArgc)
  {
    csAddr = lpszArgv[i1++];  // the IP address (we hope)

    if(i1 < (int)dwArgc) // the time zone (we hope)
    {
      csTZ = lpszArgv[i1];
    }
  }
  else
  {
    csAddr = "time.nist.gov:13";  // my default time web source, atomic clock in boulder CO
    csTZ = "UTC";
  }

  // default port is port 13.  Default access is through LAN or dial-up

  // main message/event loop

  CDateTimeThread *pThread = (CDateTimeThread *)AfxBeginThread(RUNTIME_CLASS(CDateTimeThread),
                                                               THREAD_PRIORITY_NORMAL - 1,
                                                               0, CREATE_SUSPENDED, NULL);

  if(!pThread)
  {
    AddToMessageLog("Unable to start monitor thread in 'ServiceStart()'");

    ReportStatusToSCMgr(SERVICE_STOP_PENDING, dwLastErr, 0);

    return;
  }

  pThread->m_csSocks = csSocks;
  pThread->m_csAddr = csAddr;
  pThread->m_csTZ = csTZ;
  pThread->m_bDebug = bDebug;  // for DEBUG mode it behaves differently...

  HANDLE aHandles[2];  // increase size as needed for more objects
  aHandles[0] = event.m_hObject;
  aHandles[1] = pThread->m_hThread;  // thread handle (if it quits, *I* quit)

  event.ResetEvent();       // must be 'reset'
  pThread->ResumeThread();  // starts the thread


  // Tell the SCM that I'm up and running now!
  ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0);

  while(1)  // this is where "the work" is done
  {
    DWORD dwWait = MsgWaitForMultipleObjects(sizeof(aHandles)/sizeof(*aHandles),
                                             aHandles, FALSE, 100,
                                             QS_ALLEVENTS | QS_SENDMESSAGE);

    if(dwWait == WAIT_OBJECT_0)
    {
      pThread->PostThreadMessage(WM_QUIT, 0, 0);
      WaitForSingleObject(aHandles[1], 3000);  // wait 3 sec for termination

      if(bDebug)
      {
        AddToMessageLog("Received WM_QUIT message - process terminating now.");
      }

      break;
    }
    else if(dwWait == (WAIT_OBJECT_0 + 1))  // the thread
    {
      AddToMessageLog("Main thread ended - service terminating on error.");

      break;
    }
    else if(dwWait != WAIT_TIMEOUT &&  // neither a timeout nor a message
            dwWait != (WAIT_OBJECT_0 + sizeof(aHandles)/sizeof(*aHandles)))
    {
      CString csTemp;
      csTemp.Format("ABNORMAL TERMINATION of 'MsgWaitForMultipleObjects()' - %08xH (%d)",
                    dwWait, dwWait);

      AddToMessageLog(csTemp);

      break;
    }

    // while messages are available invoke theApp.MyMessageLoop
    // This *also* handles the idle loop processing for MFC

    if(theApp.MyMessageLoop())
      break;
  }
}

VOID ServiceStop()
{
  // NOTE:  sample code suggests this should never take more than 3 seconds to complete

  ReportStatusToSCMgr(SERVICE_STOP_PENDING, dwLastErr, 0);

  if(bDebug)  // only when debugging....
  {
    while(!PostThreadMessage(dwThreadId, WM_QUIT, 0, 0))  // 'WM_QUIT' to thread
    {
      if(GetLastError() == ERROR_INVALID_THREAD_ID)
        break;
    } // keep trying until it succeeds
  }
  else
  {
    CEvent event(FALSE, FALSE, SZGLOBALEVENTNAME);

    event.SetEvent();  // that oughta do it!
  }
}

